/* ---------------------------------------------------------- 
%   Copyright (C) 1995 Takashi Chikayama
%       (Read COPYRIGHT for detailed information.) 
----------------------------------------------------------- */

/*
  Goal Object
*/

#include <klic/gdobject.h>  /* fatal */
#include "goalobj.h"
#include "atom.h"
#include "funct.h"

#define CLASS_NAME  goal
#define G_OBJ_TYPE  struct goal_object
#define G_OBJ_SIZE  G_SIZE_IN_Q(G_OBJ_TYPE)

#include <klic/gd_macro.h>
#include <klic/gmodule.h>
#include <klic/predinfo.h>
#include <klic/g_vector.h>
#include <klic/generic.h>  /* isclass */
#include "trace.h"  /* untrace_goal */

#define HashFunc(pred)	((((unsigned long) (pred)) >> 2) & hash_mask)

extern const struct predicate step_trigger_pred;
extern const struct modinfo defined_modules[];

extern struct data_object_method_table predicate_g_method_table;
extern struct data_object_method_table vector_g_method_table;

extern unsigned long enter_atom_body();
extern q size_of_vector();

/* Predicate Table */

struct pred_tab_ent {
  const struct predicate* pred;
  q modname;			/* module name atom */
  q name;			/* predicate name atom */
};

static struct pred_tab_ent* pred_tab;
static unsigned long hash_mask;

static void
make_pred_tab()
{
  declare_globals;
  int k, n_pred;
  const struct modinfo* dm;

  /* Let's first count the number predicates in this program */
  n_pred = 0;
  for (dm = &defined_modules[0]; dm->name != 0; dm++) {
    const struct predinfo* const* dp;
    for (dp = &dm->preds[0]; *dp != 0; dp++) {
      n_pred++;
    }
  }

  /* Allocate predicate table and its hash table */
  {
    unsigned long hash_size = 1;
    while (hash_size <= (3*n_pred/2)) hash_size *= 2;
    pred_tab =
      (struct pred_tab_ent*) calloc(hash_size, sizeof(struct pred_tab_ent));
    hash_mask = hash_size - 1;
    for (k = 0; k < hash_size; k++) {
      pred_tab[k].pred = 0;
    }
  }

  /* fill predicate table */
  for (dm = &defined_modules[0]; dm->name != 0; dm++) {
    q modname = makesym(enter_atom_body(dm->name, strlen(dm->name)));
    const struct predinfo* const* dp;
    for (dp = &dm->preds[0]; *dp != 0; dp++) {
      const struct predicate* pred = (*dp)->predstruct;
      long h = HashFunc(pred);
      while (pred_tab[h].pred != pred) {
	if (pred_tab[h].pred == 0) {
	  pred_tab[h].pred = pred;
	  pred_tab[h].modname = modname;
	  pred_tab[h].name =
	    makesym(enter_atom_body((*dp)->name, strlen((*dp)->name)));
	  break;
	}
	h = (h + 1) & hash_mask;
      }
    }
  }
}

static struct pred_tab_ent*
find_pred_ent(pred)
  struct predicate* pred;
{
  static int pred_tab_made = 0;
  long h;
  if (!pred_tab_made) {
    make_pred_tab();
    pred_tab_made = 1;
  }
  for (h = HashFunc(pred);
       pred_tab[h].pred != pred;
       h = (h + 1) & hash_mask) {
    if (pred_tab[h].pred == NULL) {
      /* somehow we couldn't find the predicate */
      fatal("Predicate not found in the table");
    }
  }
  return &pred_tab[h];
}

/* Copy for Garbage Colleciton */

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;
  struct goalrec *goal = g_self->goal;
  struct goalrec* newgoal;
  long arity, k;

  arity = goal->pred->arity;
  newgoal = (struct goalrec*) klic_alloc(arity + 2);
  newgoal->next = 0;
  newgoal->pred = goal->pred;
  for (k = 0; k < arity; k++) {
    G_COPY_KL1_TERM_TO_NEWGEN(goal->args[k], newgoal->args[k]);
  }
  G_SET_NEWOBJ_IN_NEWGEN(newself);
  newself->goal = newgoal;
  newself->id = g_self->id;
  G_RETURN_FROM_GC(newself);
}

/* Body Methods */

GDDEF_METHOD(reduce_0)
{
  G_STD_DECL;
  resume_same_prio(g_self->goal);
}

GDDEF_METHOD(step_4)
{
  G_STD_DECL;
  struct goalrec *newgoal;
  GD_ALLOC_AREA(newgoal, (struct goalrec *), 2+5);
  newgoal->pred = &step_trigger_pred;
  newgoal->args[0] = makefunctor(g_self);
  newgoal->args[1] = g_argv[0]; /* children info */
  newgoal->args[2] = g_argv[1]; /* woken-up info */
  newgoal->args[3] = g_argv[2]; /* failure info */
  newgoal->args[4] = g_argv[3]; /* suspension reason info */
  resume_same_prio(newgoal);
}

GDDEF_METHOD(arg_2)
{
  G_STD_DECL;
  long size, position;
  struct goalrec *goal;
  q the_arg;

  goal = g_self->goal;
  size = goal->pred->arity;
  GDSET_INTARG_WITHIN_RANGE(position, g_argv[0], 1, size+1);
  the_arg = goal->args[position-1];

  do_shallow_unify(g_argv[1], the_arg);
}

GDDEF_METHOD(setarg_4)
{
  G_STD_DECL;
  long size, position;
  G_OBJ_TYPE* newobj;
  struct goalrec *oldgoal, *newgoal;
  q oldarg;
  long k;

  size = g_self->goal->pred->arity;
  GDSET_INTARG_WITHIN_RANGE(position, g_argv[0], 1, size+1);
  oldarg = g_self->goal->args[position-1];

  /* Allocate and set up new goal record */
  GD_ALLOC_AREA(newgoal, (struct goalrec *), size+2);
  oldgoal = g_self->goal;
  newgoal->pred = oldgoal->pred;
  for (k = 0; k < size; k++) {
    newgoal->args[k] = oldgoal->args[k];
  }
  newgoal->args[position-1] = g_argv[2];

  /* Allocate and set up new goal object */
  GDSET_NEWOBJ(newobj);
  newobj->goal = newgoal;

  do_shallow_unify(g_argv[1], oldarg);
  G_UNIFY_VALUE(g_argv[3], makefunctor(newobj));
}

GDDEF_METHOD(id_1)
{
  G_STD_DECL;
  unsigned long id;

  id = g_self->id;
  G_UNIFY_VALUE(g_argv[0], makeint(id));
}

GDDEF_METHOD(predicate_4)
{
  G_STD_DECL;
  struct pred_tab_ent *pte;

  pte = find_pred_ent(g_self->goal->pred);
  G_UNIFY_VALUE(g_argv[0], pte->modname);
  G_UNIFY_VALUE(g_argv[1], pte->name);
  G_UNIFY_VALUE(g_argv[2], makeint(pte->pred->arity));
  G_UNIFY_VALUE(g_argv[3], makeint((unsigned long) pte->pred >> 2));
}

GDDEF_METHOD(set__predicate_2)
{
  G_STD_DECL;
  long arity;
  G_OBJ_TYPE* newobj;
  struct goalrec *oldgoal, *newgoal;
  q predq;
  const struct predicate_object* pred;
  long k;

  oldgoal = g_self->goal;
  arity = oldgoal->pred->arity;

  predq = g_argv[0];
  GD_DEREF(predq);
  if( !G_ISGOBJ(predq) || !isclass(predq, predicate) ){
    GD_ERROR_IN_METHOD("First argument is not a predicate object",
		       "set_predicate");
  }
  pred = (const struct predicate_object*) functorp(predq);
  if (pred->pdesc->arity != arity) {
    GD_ERROR_IN_METHOD("Arity of given predicate object doesn't agree with the goal",
		       "set_predicate");
  }

  /* Allocate and set up new goal record */
  GD_ALLOC_AREA(newgoal, (struct goalrec *), arity+2);
  newgoal->pred = pred->pdesc;
  for (k = 0; k < arity; k++) {
    newgoal->args[k] = oldgoal->args[k];
  }

  /* Allocate and set up new goal object */
  GDSET_NEWOBJ(newobj);
  newobj->goal = newgoal;

  G_UNIFY_VALUE(g_argv[1], makefunctor(newobj));
}

/*  Generic Method Table */
GDDEF_GENERIC()
{
  G_STD_DECL;

  switch( g_method_functor ){
    GD_METHOD_CASE(reduce_0);
    GD_METHOD_CASE(step_4);
    GD_METHOD_CASE(arg_2);
    GD_METHOD_CASE(setarg_4);
    GD_METHOD_CASE(predicate_4);
    GD_METHOD_CASE(id_1);
    GD_METHOD_CASE(set__predicate_2);
    GD_METHOD_CASE_DEFAULT;
  }
}

/* Guard Methods */

GDDEF_GGENERIC()
{
  G_STD_DECL;
  switch( g_method_functor ){
    GD_GMETHOD_CASE_DEFAULT;
  }
}

/* Printer is not specialized for now */

/* Comparison is not defined */

/* Hashing is not defined */

#define GDUSE_MY_GC
#define GDUSE_MY_GENERIC
#define GDUSE_MY_GGENERIC

/* Define the method table structure */

#include <klic/gd_methtab.h>

/* Object Creation */

static long goal_obj_id = 0;

extern q
make_goal_object(goal)
  struct goalrec* goal;
{
  declare_globals;
  struct goal_object* obj = (struct goal_object*)
    klic_alloc((sizeof(struct goal_object) + sizeof(long) - 1) / sizeof(long));
  untrace_goal(goal);
  obj->method_table = &goal_g_method_table;
  obj->goal = goal;
  obj->id = goal_obj_id++;
  return makefunctor(obj);
}

G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* newobj;
  struct goalrec *newgoal;
  q predq, argvq;
  struct predicate_object *pred;
  long pred_arity, argv_arity;
  long k;

  if (g_argc != 2) {
    G_ERROR_IN_NEW("Wrong number of arguments");
  }
  predq = g_argv[0];
  G_DEREF_FOR_NEW(predq);
  if( !G_ISGOBJ(predq) || !isclass(predq, predicate) ){
    G_ERROR_IN_NEW("First argument is not a predicate object");
  }
  pred = (struct predicate_object *)functorp(predq);
  pred_arity = pred->pdesc->arity;

  argvq = g_argv[1];
  G_DEREF_FOR_NEW(argvq);
  if( !G_ISGOBJ(argvq) || !isclass(argvq, vector) ){
    G_ERROR_IN_NEW("Second argument is not a vector object");
  }
  argv_arity = intval(size_of_vector(argvq));

  if (pred_arity != argv_arity) {
    G_ERROR_IN_NEW("Number of args mismatch with the predicate");
  }

  GD_ALLOC_AREA_FOR_NEW(newgoal, (struct goalrec *), pred_arity+2);
  newgoal->pred = pred->pdesc;
  for (k = 0; k < pred_arity; k++) {
    extern q element_of_vector();
    newgoal->args[k] = element_of_vector(argvq, makeint(k));
  }
  G_SET_NEWOBJ_FOR_NEW(newobj, G_SIZE_IN_Q(G_OBJ_TYPE));
  newobj->goal = newgoal;
  newobj->id = goal_obj_id++;
  return makefunctor(newobj);
}
